#include "config.h"

#include <fstream>

#include "json5cpp.h"
#include "log.h"

namespace Utils {
bool ends_with(const std::string &str, const std::string &suffix) {
  return str.size() >= suffix.size() &&
         str.compare(str.size() - suffix.size(), suffix.size(), suffix) == 0;
}
}  // namespace Utils

template <>
int Config::Parse<Config::ConfigFormat::JSON5>(std::ifstream &input) {
  Json::Value value;
  std::string err;
  if (Json5::parse(input, value, &err)) {
    dict_.Import(value);
  } else {
    LOG(ERROR) << "Invalid JSON5: " << err;
    return -1;
  }
  return 0;
}

template <>
int Config::Parse<Config::ConfigFormat::CONF>(std::ifstream &in) {
  bool valid = true;

  std::string line;
  while (!std::getline(in, line).eof()) {
    // Perform left trim
    line.erase(line.begin(),
               std::find_if(line.begin(), line.end(), [](int ch) { return !std::isspace(ch); }));

    // Skip empty lines
    if (line.empty()) {
      continue;
    }

    // Skip comments
    if (line.size() > 1 && line[0] == '#') {
      continue;
    }

    auto delim = line.find_first_of(':');
    if (delim == std::string::npos) {
      // Skip ill-formatted lines
      continue;
    }

    std::string key = line.substr(0, delim);
    std::string value = line.substr(delim + 1);

    if (key.empty()) {
      // Skip ill-formatted lines
      continue;
    }

    if (value.empty()) {
      LOG(ERROR) << "Configuration entry " << key << " has no value!";
      valid &= false;
    }

    dict_.Insert(key, value);
  }

  if (!valid) {
    throw std::invalid_argument("Configuration error: Keys with no value detected!");
  }

  return 0;
}

//Config::SetupConfig::SetLogSeverity(const std::string &s) {
//  if (s == "TRACE") min_log_severity_(TRACE);
//  if (s == "DEBUG") min_log_severity_(DEBUG);
//  if (s == "INFO") min_log_severity_(INFO);
//  if (s == "WARNING") min_log_severity_(WARNING);
//  if (s == "ERROR") min_log_severity_(ERROR);
//   if (s == "FATAL") min_log_severity_(FATAL);
//   throw std::invalid_argument("Configuration error: Unknown log_min_severity '" + s + "'");
// }

Config::Config(const std::string &filename) {
  // Open file and check if valid
  std::ifstream in(filename.c_str(), std::ios_base::in);
  if (!in.is_open()) {
    throw std::invalid_argument("Configuration file (" + filename +
                                ") could not be opened. Does it exist?");
  }

  // Pass open file handle to parser according to filename ending
  if (Utils::ends_with(filename, ".json5")) {
    Parse<ConfigFormat::JSON5>(in);
  } else if (Utils::ends_with(filename, ".conf")) {
    Parse<ConfigFormat::CONF>(in);
  } else {
    throw std::invalid_argument("Configuration file (" + filename +
                                ") has unrecognized suffix. Please specify the intended format.");
  }

  setup_configs_ = std::make_shared<SetupConfig>(dict_);
  run_configs_ = std::make_shared<RunConfig>(dict_);
}

void Config::print() const { dict_.Print(); }

//uint32_t Config::getSourceID() const {
//  uint32_t source_id = 0;
//  if (getProcessorType() == ProcessorType::GMT) {
//    source_id = 1;
//  } else if (getProcessorType() == ProcessorType::CALO) {
//    source_id = 2;
//  } else {
//    if (getCMSSWHeaders()) {
//      LOG(ERROR) << "PROCESSOR_TYPE INCOMPATIBLE WITH CMSSW HEADER OUTPUT OPTION, PLEASE DISABLE "
//                    "CMSSW HEADERS OR CHANGE PROCESSOR_TYPE, EXITING";
//      throw std::invalid_argument(
//          "ERROR: PROCESSOR_TYPE INCOMPATIBLE WITH CMSSW HEADER OUTPUT OPTION, PLEASE DISABLE "
//          "CMSSW HEADERS OR CHANGE PROCESSOR_TYPE");
//    } else {
//      LOG(WARNING) << "Undefined source ID has been requested: returning 0.";
//    }
//  }
//  return source_id;
//}