#ifndef CONFIG_H
#define CONFIG_H

#include <jsoncpp/json/json.h>

#include <boost/lexical_cast.hpp>
#include <cstdint>
#include <map>
#include <stdexcept>
#include <string>

#include "log.h"
#include "processor.h"
#include "utils.h"

class ConfigMap {
 public:
  using ValueType = Json::Value;

 private:
  ValueType data_;

 public:
  void Insert(const std::string &key, ValueType val) { data_[key] = std::move(val); }

  void Import(const ValueType &values) { data_ = values; }

  void Print() const {
    for (auto it = data_.begin(); it != data_.end(); it++) {
      auto k = it.key().asString();
      LOG(INFO) << "key " << k << " | value" << data_[k];
    }
  }

  std::string at(const std::string &key) const {
    try {
      auto &val = data_[key];
      return val.asString();
    } catch (std::exception &e) {
      LOG(ERROR) << "Error fetching value for key '" << key << "'";
      LOG(ERROR) << e.what();
      throw;
    }
  }

  ValueType get(const std::string &key) const {
    try {
      auto &val = data_[key];
      return val.asString();
    } catch (std::exception &e) {
      LOG(ERROR) << "Error fetching value for key '" << key << "'";
      LOG(ERROR) << e.what();
      throw;
    }
  }
};

  class SetupConfig {
   public:
    enum class InputType { WZDMA, FILEDMA, TCPIP, MICRONDMA, FILE };
    Utils::EnumDictionary<InputType> GetInputType = {{"WZDMA", InputType::WZDMA},
                                                     {"FILEDMA", InputType::FILEDMA},
                                                     {"TCPIP", InputType::TCPIP},
                                                     {"MICRONDMA", InputType::MICRONDMA},
                                                     {"FILE", InputType::FILE}};

    struct SconeAddress {
      std::string host;
      std::string port;
      std::string board;

      SconeAddress(std::string h, std::string p, std::string b) : host(h), port(p), board(b) {}

      std::string Get() { return host + ":" + port + "/" + board; }
    };

    explicit SetupConfig(ConfigMap &dict)
      : scone_address_({dict.at("scone_host"), dict.at("scone_port"), dict.at("scone_board")}),
        tcp_dest_ports_range_(dict.at("tcpDestPort")),
        dma_device_(dict.at("dma_dev")),
        attempts_auto_reconnect_(dict.at("dev_TCPAutoReconnectOnFailure") == "true"),
        source_id_(dict.get("sourceID").asUInt()),
        port_(dict.get("port").asUInt()),
        num_packets_per_report_(dict.get("packets_per_report").asUInt()),
        num_packet_buffers_(dict.get("dma_number_of_packet_buffers").asUInt64()),
        packet_buffer_size_(dict.get("dma_packet_buffer_size").asUInt64()),
        num_threads_(dict.get("threads").asUInt()),
        max_output_size_(dict.get("max_file_size").asUInt64())
  {
    std::string severity = dict.at("log_min_severity");
    min_log_severity_ = GetLogLevel(severity);
    input_type_ = GetInputType(dict.at("input"));
  }

  InputType input_type_;
  SconeAddress scone_address_;
  std::string tcp_dest_ports_range_;
  std::string dma_device_;
  bool attempts_auto_reconnect_;
  uint32_t port_;
  uint32_t source_id_;

  uint32_t num_threads_;
  uint64_t max_output_size_;

  // Statistics
  uint32_t num_packets_per_report_;
  uint64_t num_packet_buffers_;
  uint64_t packet_buffer_size_;

  LOG_LEVEL min_log_severity_;
};

class RunConfig {
 public:
  explicit RunConfig(ConfigMap &dict)
      : run_id_(dict.get("run_number").asInt()),
        // Input
        num_orbits_per_packet_(dict.get("nOrbitsPerPacket").asUInt64()),
        // Processing
        processor_type_(static_cast<ProcessorType>(GetProcessorType(dict.at("processorType")))),
        enable_processing_(dict.at("enable_stream_processor") == "yes"),
        apply_zero_suppression_(dict.at("doZS") == "yes"),

        // Output
        output_file_base_(dict.at("output_filename_base")),
        output_file_prefix_(dict.at("output_filename_prefix")),
        force_write_out_(dict.at("output_force_write") == "yes"),
        num_orbits_per_file_(dict.get("nOrbitsPerFile").asUInt64()) {}

  // Input
  std::vector<std::string> input_ports_;
  std::vector<std::string> input_files_;
//  input_files_(dict.at("input_file").asString()), // should be multiple files in case of N streams
//      input_ports_(dict.at("tcpPorts")), // N streams
  uint64_t num_orbits_per_packet_;
  uint32_t input_port_;
  std::string input_file_;


  // Processor
  ProcessorType processor_type_;
  bool enable_processing_;
  bool apply_zero_suppression_;


  // Output
  std::string output_file_base_;
  std::string output_file_prefix_;

  uint32_t run_id_; // From control structure
  bool force_write_out_;

  // File Stream outputs
  uint64_t max_file_size_;

  // Orbit Stream outputs
  bool packs_cmssw_headers_;
  uint64_t num_orbits_per_file_;
};

class Config {
 public:
  enum class ConfigFormat { CONF, JSON5 };
  explicit Config(const std::string &filename);

  template <ConfigFormat>
  int Parse(std::ifstream &input);

  void print() const;

  std::shared_ptr<SetupConfig> setup_configs_;
  std::shared_ptr<RunConfig> run_configs_;

 private:
  ConfigMap dict_;
};

//class InputConfig : public Config {
// public:
//  explicit InputConfig(SetupData &setup_data, RunConfig &run_data)
//  {}
//  int packet_buffer_size;
//  int num_packet_buffers;
//  int packets_per_report;
//
//  // Special variables
//  std::string input_file;
//  SetupData::SconeAddr scone_address;
//};
//
//
//class ProcessorConfig : public Config {
// public:
//  explicit ProcessorConfig() = default();
//  const bool is_active;
//  const int run_number;
//  const bool apply_zero_suppression;
//};
//    ProcessorType getProcessorType() const {
//      const std::string &input = dict_.at("processor_type");
//      if (input == "PASS_THROUGH") {
//        return ProcessorType::PASS_THROUGH;
//      }
//      if (input == "GMT") {
//        return ProcessorType::GMT;
//      }
//      if (input == "CALO") {
//        return ProcessorType::CALO;
//      }
//      if (input == "BRIL") {
//        return ProcessorType::BRIL;
//      }
//      throw std::invalid_argument("Configuration error: Wrong processor type '" + input + "'");
//    }
//
//    bool getDoZS() const { return (dict_.at("doZS") == "yes"); }
//
//    uint32_t getNumThreads() const { return boost::lexical_cast<uint32_t>(dict_.at("threads")); }
//
//
//  };

//  class OutputConfig {
//    std::string getOutputFilenamePrefix() const { return dict_.at("output_filename_prefix"); }
//
//    std::string getOutputFilenameBase() const { return dict_.at("output_filename_base"); }
//
//    uint64_t getOutputMaxFileSize() const {
//      return boost::lexical_cast<uint64_t>(dict_.at("max_file_size"));
//    }
//
//    bool getOutputForceWrite() const { return (dict_.at("output_force_write") == "yes"); }
//  };

#endif
