#include "pipeline.h"

Pipeline::Pipeline(Config &conf, uint maxTokensPerThread)
    : fMaxTokensPerThread(maxTokensPerThread),
      conf_(conf),
      fStreamProcessor(conf.run_configs_->apply_zero_suppression_, conf.run_configs_->processor_type_,
                       conf.run_configs_->num_orbits_per_packet_,
                       conf.run_configs_->packs_cmssw_headers_,
                       static_cast<uint16_t>(conf.setup_configs_->source_id_), control) {
  tools::log::set_filter_min_severity(conf_.setup_configs_->min_log_severity_);
  conf_.print();
  const SetupConfig::InputType &input = conf_.setup_configs_->input_type_;
  const size_t &nbPacketBuffers = conf_.setup_configs_->num_packet_buffers_;
  const size_t &packetBufferSize = conf_.setup_configs_->packet_buffer_size_;

  if (input == SetupConfig::InputType::FILEDMA) {
    // Create FILE DMA reader
    fInputFilter = std::make_shared<FileDmaInputFilter>(conf_.run_configs_->input_file_,
                                                        packetBufferSize, nbPacketBuffers, control);

  } else if (input == SetupConfig::InputType::WZDMA) {
    // Create WZ DMA reader
    fInputFilter =
        std::make_shared<WZDmaInputFilter>(packetBufferSize, nbPacketBuffers, control, conf_);

  } else if (input == SetupConfig::InputType::MICRONDMA) {
    // create MicronDmaInputFilter reader
    fInputFilter =
        std::make_shared<MicronDmaInputFilter>(packetBufferSize, nbPacketBuffers, control, conf_);

  } else if (input == SetupConfig::InputType::TCPIP) {
    // create TcpInputFilter reader
    if (conf_.setup_configs_->attempts_auto_reconnect_) {
      fInputFilter = std::make_shared<TcpInputFilter>(
          conf_.run_configs_->input_port_, packetBufferSize, nbPacketBuffers, conf_.run_configs_->num_orbits_per_packet_, control);

    } else {
      throw std::invalid_argument(
          "Configuration error: enable developmentMode to use TCP input filter");
    }
  } else {
    throw std::invalid_argument("Configuration error: Unknown input type was specified");
  }

  // Add input reader to a pipeline
  fPipeline.add_filter(*fInputFilter);

  // Create reformatter and add it to the pipeline
  if (conf_.run_configs_->enable_processing_) {
    fPipeline.add_filter(fStreamProcessor);
  }

  // Create file-writing stage and add it to the pipeline
  if (conf_.run_configs_->num_orbits_per_file_) {
    fOutputStream = std::make_shared<OutputByOrbitStream>(control, conf_);
  } else {
    fOutputStream = std::make_shared<OutputBySizeStream>(
        conf_.run_configs_->output_file_base_, conf_.run_configs_->output_file_prefix_,
        Pipeline::control, conf_.run_configs_->processor_type_);
  }
  fPipeline.add_filter(*fOutputStream);

  //    tbb::tick_count mainStartTime = tbb::tick_count::now();
  control.running = false;
  control.run_number = 0;
  control.max_file_size = conf_.setup_configs_->max_output_size_;  // in Bytes
  control.packets_per_report = static_cast<int>(conf_.setup_configs_->num_packets_per_report_);
  control.output_force_write = conf_.run_configs_->force_write_out_;
  control.n_orbits_per_packet = static_cast<int>(conf_.run_configs_->num_orbits_per_packet_);

  // Firmware needs at least 1MB buffer for DMA
  if (conf_.setup_configs_->packet_buffer_size_ < 1024 * 1024) {
    LOG(ERROR) << "dma_packet_buffer_size must be at least 1048576 bytes (1MB), but "
               << conf_.setup_configs_->packet_buffer_size_
               << " bytes was given. Check the configuration file.";
    throw std::invalid_argument("Configuration error: DMA packet buffer size too small. ");
  }
}

MultiPipeline::MultiPipeline(std::vector<Config> &configs) {
  for (auto &config : configs) {
    try {
      fPipelines.emplace_back(config);
      LOG(DEBUG) << "-Configuration loaded";
      LOG(DEBUG) << "-Pipeline set up";
    } catch (std::invalid_argument &e) {
      LOG(FATAL) << "-Configuration invalid! Error text is \"" << e.what() << "\" Bailing out.";
      throw;
    } catch (std::exception &e) {
      LOG(ERROR) << "-Error occurred. error text is: \"" << e.what() << "\"";
      throw;
    }
  }

  {  // RAII
    fServer = std::unique_ptr<server>(
        new server(io_service, fPipelines[0].conf_.setup_configs_->port_, Pipeline::control));
    fServerThread = boost::thread(boost::bind(&boost::asio::io_service::run, &io_service));
  }
}

int MultiPipeline::Run() {
  int retval = 0;
  assert(!fPipelines.empty());
  try {
    if (!RunPipelines(static_cast<unsigned int>(fPipelines.size()))) {
      retval = 1;  // Will terminate with error code set
    }
  } catch (std::exception &e) {
    LOG(ERROR) << "Error in pipelines. Error text is: \"" << e.what() << "\"";
    retval = 1;                             // Will terminate with error code set
    Config &conf = fPipelines.at(0).conf_;  // Use info from first file for SCONE reset
    SetupConfig::SconeAddress &scone = conf.setup_configs_->scone_address_;
    LOG(INFO) << "Resetting board at http://" << scone.Get() << " before exiting.. ";
    int res = tools::reset_board(scone.host, scone.port, scone.board);
    if (res != 0) {
      LOG(ERROR) << "Reset failed! Error code: " << res;
    } else {
      LOG(INFO) << "Successfully reset the board.";
    }
  }

  LOG(DEBUG) << "Terminating the internal TCP server.";
  io_service.stop();
  if (!fServerThread.try_join_for(boost::chrono::milliseconds(2000))) {
    LOG(ERROR) << "Failed to terminate the internal TCP server, the program "
                  "may crash on exit.";
  } else {
    LOG(DEBUG) << "Internal TCP server is terminated.";
  }

  return retval;
}

int MultiPipeline::RunPipelines(uint numStreams) {
  tbb::task_group group;
  // Run pipelines and measure real time
  tbb::tick_count t0 = tbb::tick_count::now();
  for (uint i = 0; i < numStreams; ++i) {
    auto &p = this->fPipelines.at(i);
    group.run([&] { return p.fPipeline.run(p.conf_.setup_configs_->num_threads_ * p.fMaxTokensPerThread); });
  }
  group.wait();
  tbb::tick_count t1 = tbb::tick_count::now();

  if (fVerbose) {
    LOG(INFO) << "time = " << (t1 - t0).seconds();
  }
  return 1;
}