#include "processor.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <iomanip>
#include <vector>

#include "format.h"
#include "log.h"
#include "slice.h"

// Definition of the static member stats
StreamProcessor::Statistics StreamProcessor::stats;

StreamProcessor::StreamProcessor(const bool doZS_, ProcessorType processorType_,
                                 const uint32_t nOrbitsPerPacket_, const bool cmsswHeaders_,
                                 const uint16_t sourceID_, ctrl &control_)
    : tbb::filter(parallel),
      nbPackets(0),
      doZS(doZS_),
      processorType(processorType_),
      nOrbitsPerPacket(nOrbitsPerPacket_),
      cmsswHeaders(cmsswHeaders_),
      sourceID(sourceID_),
      control(control_) {
  LOG(TRACE) << "Created transform filter at " << static_cast<void *>(this);
}

BrilHistoQueue<std::array<uint32_t, constants::NBXPerOrbit + constants::NFramesInHistoHeader>>
    StreamProcessor::BrilQueue;

StreamProcessor::~StreamProcessor() = default;

// checks that the packet size is an integer multiple of the BX block size,
// minus the header/trailers
bool StreamProcessor::CheckFrameMultBlock(size_t inputSize) const {
  int bsize = sizeof(blockMuon);
  if ((inputSize - nOrbitsPerPacket * constants::orbit_trailer_size - 32 * nOrbitsPerPacket - 32) %
          bsize !=
      0) {
    stats.n_consistent_sized_packets = 0;
    stats.packet_skipped_inconsistent_size++;

    if ((stats.packet_skipped_inconsistent_size++ == 1) ||
        ((stats.packet_skipped_inconsistent_size < 100) &&
         (stats.packet_skipped_inconsistent_size % 10 == 0)) ||
        ((stats.packet_skipped_inconsistent_size < 1000) &&
         (stats.packet_skipped_inconsistent_size % 100 == 0)) ||
        ((stats.packet_skipped_inconsistent_size < 10000) &&
         (stats.packet_skipped_inconsistent_size % 1000 == 0)) ||
        (stats.packet_skipped_inconsistent_size % 10000 == 0)) {
      LOG(WARNING) << "Frame size not a multiple of block size after headers "
                      "and trailers have been subtracted. Counted "
                   << stats.packet_skipped_inconsistent_size << " packets skipped.";
    }

    LOG(WARNING) << "Frame size not a multiple of block size after orbit headers "
                    "(32B*nOrbitsPerPacket), orbit trailers (512B*nOrbitsPerPacket), "
                    "and packet trailer (32B) have been subtracted. \n Frame size = "
                 << inputSize << ", block size = " << bsize << ", packet will be skipped";
    return false;
  } else {
    stats.n_consistent_sized_packets++;
    if (stats.n_consistent_sized_packets == 6000000) {  // every ~10 mins
      LOG(WARNING) << "Resetting packet_skipped_inconsistent_size counter to 0 "
                      "after 6000000M consistent packets. Count was at "
                   << stats.packet_skipped_inconsistent_size;
      stats.packet_skipped_inconsistent_size = 0;
    }
  }
  return true;
}

bool StreamProcessor::GetTrailer(Slice &input, char *&rd_ptr) {
  rd_ptr += 32;  // +32 to account for orbit header
  while (rd_ptr + sizeof(orbit_trailer) - 1 <= input.end()) {
    auto *ot = reinterpret_cast<orbit_trailer *>(rd_ptr);
    if (ot->beefdead[0] == constants::beefdead) {  // found orbit trailer
      return true;
    }
    rd_ptr += sizeof(blockMuon);
  }
  return false;
}

inline std::pair<uint32_t, bool> StreamProcessor::ProcessOrbitHeader(char *rd_ptr) {
  // get orbit from orbit header
  auto *bl_pre =
      reinterpret_cast<blockMuon *>(rd_ptr);  // blockMuon.orbit is identical to blockCalo.frame0
  auto orbitN = uint32_t{bl_pre->orbit[0]};
  // save warning_test_enable bit
  auto mask = uint32_t(1) << 31;
  bool warning_test_enable = ((orbitN & mask) == mask);
  // remove warning_test_enable bit from orbit header
  orbitN &= 0x7fffffff;
  std::pair<uint32_t, bool> orbit_header = std::pair<uint32_t, bool>{orbitN, warning_test_enable};
  return orbit_header;
}

// Goes through orbit worth of data and fills the output memory with the calo
// data corresponding to the non-empty bunch-crossings, as marked in bx_vect
StreamProcessor::fillOrbitMetadata StreamProcessor::FillOrbitCalo(orbit_trailer *trailer,
                                                                  char *rd_ptr, char *wr_ptr,
                                                                  const char *rd_end_ptr,
                                                                  const char *wr_end_ptr) {
  std::pair<uint32_t, bool> orbit_header = std::pair<uint32_t, bool>{
      ProcessOrbitHeader(rd_ptr)};  //.second is the warning test enable bit
  rd_ptr += 32;                     // +32 to account for orbit header
  if (cmsswHeaders) {
    wr_ptr += sizeof(FRDFileHeader_v2);
  }  // reserving space for cmssw orbit header

  uint32_t orbit = uint32_t{orbit_header.first} -
                   1;  // Starting with orbit number one lower than what is in the header
                       // because the "link orbit" contains a few BXs of the previous orbit
  auto counts = uint32_t{0};
  uint32_t filled_bxs = 0;
  // We loop over the BX map from the orbit trailer and then match the filled
  // BXs to the data we got. The logic below is annoyingly convoluted: The first
  // BX we get in the data stream is from BX 3555, however the BX map starts at
  // BX 1, we therefore need to start reading the BX map from the 3555th bit.
  // 3555//32 = 111, so we start at the 111th word; 3555 mod 32 = 3, however we
  // start counting at BX1, so the start bit is 2.
  uint32_t bx = 3554;  // We start at 3554 here, because we increment by 1
                       // immediately after starting the loops.
  size_t word = 111;
  size_t bit = 1;      // Will be incremented immediately after starting the loop
  for (size_t pseudo_bx = 0; pseudo_bx < 3564;
       ++pseudo_bx) {  // Looping over at most an entire orbit here.
    if (word == 14 * 8 - 1 && bit == 11) {  // Bit 11 in word 111 is the last valid BX (3564), so we
                                            // roll over afterwards. (== 11 because that's the one
                                            // we worked on in the previous loop iteration)
      word = 0;
      bit = 0;
      bx = 0;  // Will be immediately incremented to 1.
      ++orbit;
    } else if (bit < 31) {
      ++bit;
    } else {
      bit = 0;
      ++word;
    }
    ++bx;
    if ((trailer->bx_map[word] & (1 << bit)) == 0) {
      continue;  // If the bit is zero that BX was empty and we continue.
    }
    assert(rd_ptr + sizeof(blockCalo) - 1 <= rd_end_ptr);
    ++filled_bxs;
    auto *bl = reinterpret_cast<blockCalo *>(rd_ptr);
    if (bl->calo0[0] == constants::beefdead) {
      break;
    }  // orbit trailer has been reached, end of orbit data
    assert(wr_ptr + (3 + 8 * 7) * 4 - 1 <= wr_end_ptr);
    auto header = uint32_t{orbit_header.second};  // header can be added to later
    memcpy(wr_ptr, (char *)&header, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&bx, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&orbit, 4);
    wr_ptr += 4;
    for (uint32_t i = 0; i < 8; i++) {
      memcpy(wr_ptr, (char *)&i, 4);
      wr_ptr += 4;  // gives link number, can later be used to find object type
      memcpy(wr_ptr, (char *)&bl->calo0[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo1[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo2[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo3[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo4[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo5[i], 4);
      wr_ptr += 4;
    }
    counts += 1;
    rd_ptr += sizeof(blockCalo);
  }

  StreamProcessor::fillOrbitMetadata meta = {counts, orbit, filled_bxs};
  return meta;
}
// add to header. allocate new[]
size_t StreamProcessor::fillFRDEventHeader_V6(char *wr_ptr_FRDHead, uint32_t inputSize,
                                              uint32_t orbit) const {
  uint16_t header_version = 6;
  uint16_t flags = 0;
  uint32_t crc_dummy = 0;
  uint32_t source_id = sourceID;  // extra 4 byte source ID used for scouting, not strictly part of
                                  // the FRDEventHeader_V6
  uint32_t eventSize = inputSize + static_cast<uint32_t>(sizeof(source_id));

  uint32_t lumisection = 1 + static_cast<uint32_t>(orbit / constants::N_orbits_per_lumisection);
  FRDEventHeader_V6 frdEventHeader_V6(header_version, flags, control.run_number, lumisection, orbit,
                                      eventSize, crc_dummy);
  memcpy(wr_ptr_FRDHead, (char *)&(frdEventHeader_V6), sizeof(frdEventHeader_V6));
  wr_ptr_FRDHead += sizeof(frdEventHeader_V6);
  memcpy(wr_ptr_FRDHead, (char *)&(source_id), sizeof(source_id));
  wr_ptr_FRDHead += sizeof(source_id);  // not strictly needed

  return (sizeof(frdEventHeader_V6) + sizeof(source_id));
}

uint32_t StreamProcessor::FillBril(char *rd_ptr, char *wr_ptr, const char *end_ptr) const {
  std::array<uint32_t, constants::NBXPerOrbit + constants::NFramesInHistoHeader> histo_arr;

  while (rd_ptr != end_ptr) {
    assert(rd_ptr + sizeof(brilFrame) - 1 <= end_ptr);

    auto *fr = reinterpret_cast<brilFrame *>(rd_ptr);

    if (fr->counter < constants::NFramesInHistoHeader) {
      if (fr->counter == 5) {  // add run number, 6th frame in bril header is run number
        histo_arr[fr->counter] = control.run_number & 0xffffffff;
      } else {
        histo_arr[fr->counter] = (fr->word >> 0) & 0xffffffff;
      }

    } else {
      histo_arr[(fr->counter * 2) - constants::NFramesInHistoHeader] = (fr->word >> 0) & 0xffff;

      histo_arr[(fr->counter * 2) + 1 - constants::NFramesInHistoHeader] =
          (fr->word >> 16) & 0xffff;
    }
    rd_ptr += sizeof(brilFrame);
  }
  constexpr auto packed_size = static_cast<uint32_t>(sizeof(uint32_t)) *
                               (constants::NBXPerOrbit + constants::NFramesInHistoHeader);
  memcpy(wr_ptr, (char *)&histo_arr, packed_size);
  BrilQueue.push(histo_arr);

  return packed_size;
}

// Goes through orbit worth of data and fills the output memory with the muons
// corresponding to the non-empty bunch-crossings, as marked in bx_vect
StreamProcessor::fillOrbitMetadata StreamProcessor::FillOrbitMuon(orbit_trailer *trailer,
                                                                  char *rd_ptr, char *wr_ptr,
                                                                  const char *rd_end_ptr,
                                                                  const char *wr_end_ptr) {
  std::pair<uint32_t, bool> orbit_header = std::pair<uint32_t, bool>{
      ProcessOrbitHeader(rd_ptr)};  //.second is the warning test enable bit
  rd_ptr += 32;                     // +32 to account for orbit header
  if (cmsswHeaders) {
    wr_ptr += 28;
  }  // reserving space for cmssw orbit header

  uint32_t orbit = uint32_t{orbit_header.first} -
                   1;  // Starting with orbit number one lower than what is in the header
                       // because the "link orbit" contains a few BXs of the previous orbit
  auto counts = uint32_t{0};
  uint32_t filled_bxs = 0;
  // We loop over the BX map from the orbit trailer and then match the filled
  // BXs to the data we got. The logic below is annoyingly convoluted: The first
  // BX we get in the data stream is from BX 3555, however the BX map starts at
  // BX 1, we therefore need to start reading the BX map from the 3555th bit.
  // 3555//32 = 111, so we start at the 111th word; 3555 mod 32 = 3, however we
  // start counting at BX1, so the start bit is 2.
  uint32_t bx = 3554;  // We start at 3554 here, because we increment by 1
                       // immediately after starting the loops.
  size_t word = 111;
  size_t bit = 1;      // Will be incremented immediately after starting the loop
  for (size_t pseudo_bx = 0; pseudo_bx < 3564;
       ++pseudo_bx) {  // Looping over at most an entire orbit here.
    if (word == 14 * 8 - 1 && bit == 11) {  // Bit 11 in word 111 is the last valid BX (3564), so we
                                            // roll over afterwards. (== 11 because that's the one
                                            // we worked on in the previous loop iteration)
      word = 0;
      bit = 0;
      bx = 0;  // Will be immediately incremented to 1.
      ++orbit;
    } else if (bit < 31) {
      ++bit;
    } else {
      bit = 0;
      ++word;
    }
    ++bx;
    if ((trailer->bx_map[word] & (1 << bit)) == 0) {
      continue;  // If the bit is zero that BX was empty and we continue.
    }
    assert(rd_ptr + sizeof(blockMuon) - 1 <= rd_end_ptr);
    ++filled_bxs;
    auto *bl = reinterpret_cast<blockMuon *>(rd_ptr);
    if (bl->orbit[0] == constants::beefdead) {
      break;
    }                    // orbit trailer has been reached, end of orbit data
    assert(wr_ptr + (3 + 2 * 8 * 3) * 4 - 1 <=
           wr_end_ptr);  // Assuming the maximum size a decoded muon block can use.
    int mAcount = 0;
    int mBcount = 0;
    bool AblocksOn[8];
    bool BblocksOn[8];
    for (unsigned int i = 0; i < 8; i++) {
      auto pt = uint32_t{(bl->mu1f[i] >> shifts::pt) & masks::pt};

      AblocksOn[i] = ((pt > 0) || (doZS == 0));
      if ((pt > 0) || (doZS == 0)) {
        mAcount++;
      }
      pt = (bl->mu2f[i] >> shifts::pt) & masks::pt;
      BblocksOn[i] = ((pt > 0) || (doZS == 0));
      if ((pt > 0) || (doZS == 0)) {
        mBcount++;
      }
    }
    if (mAcount + mBcount == 0) {
      rd_ptr += sizeof(blockMuon);
      LOG(WARNING) << '#' << nbPackets
                   << ": Detected a bx with zero muons, this should not "
                      "happen. Packet is skipped.";
      continue;
    }

    // header word of packed muon data contains mAcount (number of muons in
    // words 3,4) and mBcount (number of muons in words 5,5), as well as the
    // warning test enable flag.
    auto header =
        uint32_t{(mAcount << 16) + ((static_cast<uint32_t>(orbit_header.second)) << 8) + mBcount};

    counts += mAcount;
    counts += mBcount;
    memcpy(wr_ptr, (char *)&header, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&bx, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&orbit, 4);
    wr_ptr += 4;
    for (unsigned int i = 0; i < 8; i++) {
      if (AblocksOn[i]) {
        memcpy(wr_ptr, (char *)&bl->mu1f[i], 4);
        wr_ptr += 4;
        memcpy(wr_ptr, (char *)&bl->mu1s[i], 4);
        wr_ptr += 4;
        // next creating mu.extra which is a copy of bl->bx with a change to the
        // first bit
        memcpy(wr_ptr, (char *)&(bl->bx[i] &= ~0x1), 4);
        wr_ptr += 4;  // set bit 0 to 0 for first muon
      }
    }

    for (unsigned int i = 0; i < 8; i++) {
      if (BblocksOn[i]) {
        memcpy(wr_ptr, (char *)&bl->mu2f[i], 4);
        wr_ptr += 4;
        memcpy(wr_ptr, (char *)&bl->mu2s[i], 4);
        wr_ptr += 4;
        // next creating mu.extra which is a copy of bl->bx with a change to the
        // first bit
        memcpy(wr_ptr, (char *)&(bl->bx[i] |= 0x1), 4);
        wr_ptr += 4;  // set bit 0 to 1 for second muon
      }
    }
    rd_ptr += sizeof(blockMuon);
  }
  StreamProcessor::fillOrbitMetadata meta = {counts, orbit, filled_bxs};
  return meta;
}

void StreamProcessor::process(Slice &input, Slice &out) {
  nbPackets++;
  stats.packet_count++;

  char *rd_ptr = input.begin();
  char *wr_ptr = out.begin();
  char *rd_end_ptr = input.end();
  char *wr_end_ptr = out.begin() + out.avail();
  uint32_t counts = 0;
  uint32_t orbit_per_packet_count = 0;
  bool firstOrbit = true;
  StreamProcessor::fillOrbitMetadata meta{0, 0, 0};
  if (processorType == ProcessorType::PASS_THROUGH) {
    memcpy(wr_ptr, rd_ptr, input.size());
    out.set_end(out.begin() + input.size());
    out.set_counts(1);
    return;
  }
  if (processorType == ProcessorType::BRIL) {
    counts = FillBril(rd_ptr, wr_ptr, rd_end_ptr);
    out.set_end(out.begin() + counts);
    out.set_counts(counts);
    return;
  }

  if (!CheckFrameMultBlock(input.size())) {
    return;
  }
  while (true) {
    uint32_t orbitCount = 0;
    char *trailer_ptr = rd_ptr;
    bool trailerFound = GetTrailer(input, trailer_ptr);
    if (!trailerFound) {
      stats.orbit_trailer_error_count++;
      LOG(WARNING) << "Orbit trailer error: orbit trailer not found before end of data "
                      "packet. Packet will be skipped. Orbit trailer error count = "
                   << stats.orbit_trailer_error_count;
      return;
    }
    auto *trailer = reinterpret_cast<orbit_trailer *>(trailer_ptr);
    size_t additional_header_size = 0;

    if (processorType == ProcessorType::GMT) {
      meta = FillOrbitMuon(trailer, rd_ptr, wr_ptr, rd_end_ptr, wr_end_ptr);
      orbitCount = meta.counts;
      ++orbit_per_packet_count;
      if (orbit_per_packet_count > nOrbitsPerPacket) {
        stats.excess_orbits_per_packet_count++;
        return;
      }
      // 12 bytes for each muon/count then 12 bytes for each bx header
      uint32_t orbit_size_bytes = meta.counts * 12 + 12 * meta.filled_bxs;

      if (cmsswHeaders) {
        additional_header_size = fillFRDEventHeader_V6(wr_ptr, orbit_size_bytes, meta.orbit);
      }

      wr_ptr += (orbit_size_bytes + additional_header_size);

    } else if (processorType == ProcessorType::CALO) {
      meta = FillOrbitCalo(trailer, rd_ptr, wr_ptr, rd_end_ptr, wr_end_ptr);
      orbitCount = meta.counts;
      ++orbit_per_packet_count;
      // size of calo packet is 4bytes*(8links*7dataWords + 3headerWords)=236
      // bytes Note 7 data words per link because we have the "link number" word
      // + 6 words from calo L2
      uint32_t orbit_size_bytes = 4 * ((8 * 7) + 3) * meta.filled_bxs;

      if (cmsswHeaders) {
        additional_header_size = fillFRDEventHeader_V6(wr_ptr, orbit_size_bytes, meta.orbit);
      }

      wr_ptr += (orbit_size_bytes + additional_header_size);

    } else {
      LOG(ERROR) << "UNKNOWN PROCESSOR_TYPE, EXITING";
      throw std::invalid_argument("ERROR: PROCESSOR_TYPE NOT RECOGNISED");
    }
    rd_ptr += 32 + meta.filled_bxs * sizeof(blockMuon) +
              constants::orbit_trailer_size;  // 32 for orbit header, + nBXs +
                                              // orbit trailer
    assert(wr_ptr <= wr_end_ptr);
    assert(rd_ptr <= rd_end_ptr);
    counts += orbitCount;
    if (firstOrbit) {
      out.set_firstOrbitN(meta.orbit);
      firstOrbit = false;
    }

    if (rd_ptr < input.end()) {
      auto *dma_trailer_word = (uint32_t *)(rd_ptr);
      if (*dma_trailer_word == constants::deadbeef) {
        out.set_end(wr_ptr);
        out.set_counts(counts);
        return;
      }

      if (orbit_per_packet_count > nOrbitsPerPacket) {
        LOG(WARNING) << "expected DMA trailer word deadbeef, found " << std::hex
                     << *dma_trailer_word << ". Orbits per packet count " << orbit_per_packet_count
                     << ", > expected, (" << nOrbitsPerPacket << ") skipping packet.";
        if (stats.excess_orbits_per_packet_count % 10000 == 0) {
          LOG(WARNING) << "count of packets with excess # orbits "
                       << stats.excess_orbits_per_packet_count;
        }
        return;
      }
    }
  }
}
void *StreamProcessor::operator()(void *item) {
  Slice &input = *static_cast<Slice *>(item);
  Slice &out = *Slice::getAllocated();
  process(input, out);
  Slice::giveAllocated(&input);

  return &out;
}
