#ifndef PROCESSOR_H
#define PROCESSOR_H

#include <tbb/pipeline.h>
#include <unistd.h>

#include <atomic>
#include <fstream>
#include <iostream>
#include <utility>
#include <vector>

#include "bril_histo.h"
#include "cmssw/FRDEventHeader_V6.h"
#include "cmssw/FRDFileHeader_v2.h"
#include "controls.h"
#include "format.h"
#include "utils.h"

class Slice;

enum class ProcessorType { BRIL, CALO, GMT, PASS_THROUGH };
const Utils::EnumDictionary<ProcessorType> GetProcessorType = {
    {"BRIL", ProcessorType::BRIL},
    {"CALO", ProcessorType::CALO},
    {"GMT", ProcessorType::GMT},
    {"PASS_THROUGH", ProcessorType::PASS_THROUGH}};

class StreamProcessor : public tbb::filter {
 public:

  static BrilHistoQueue<
      std::array<uint32_t, constants::NBXPerOrbit + constants::NFramesInHistoHeader>>
      BrilQueue;

  StreamProcessor(bool doZS_, ProcessorType processorType_, uint32_t nOrbitsPerPacket_,
                  bool cmsswHeaders_, uint16_t sourceID, ctrl &control);
  void *operator()(void *item) override;
  ~StreamProcessor() override;

 private:
  struct fillOrbitMetadata {
    uint32_t counts;
    uint32_t orbit;
    uint32_t filled_bxs;
  };
  void process(Slice &input, Slice &out);
  bool CheckFrameMultBlock(size_t inputSize) const;
  bool GetTrailer(Slice &input, char *&rd_ptr);
  inline std::pair<uint32_t, bool> ProcessOrbitHeader(char *rd_ptr);
  fillOrbitMetadata FillOrbitMuon(orbit_trailer *trailer, char *rd_ptr, char *wr_ptr,
                                  const char *rd_end_ptr, const char *wr_end_ptr);
  fillOrbitMetadata FillOrbitCalo(orbit_trailer *trailer, char *rd_ptr, char *wr_ptr,
                                  const char *rd_end_ptr, const char *wr_end_ptr);
  uint32_t FillBril(char *rd_ptr, char *wr_ptr, const char *end_ptr) const;
  uint64_t nbPackets;
  const bool doZS;
  const ProcessorType processorType;
  const uint32_t nOrbitsPerPacket;
  const bool cmsswHeaders;
  const uint16_t sourceID;
  ctrl &control;
  size_t fillFRDEventHeader_V6(char *wr_ptr_FRDHead, uint32_t inputSize, uint32_t orbit) const;

 public:
  static class Statistics {
   public:
    Statistics()
        : orbit_trailer_error_count(0),
          packet_count(0),
          excess_orbits_per_packet_count(0),
          packet_skipped_inconsistent_size(0),
          n_consistent_sized_packets(0) {}

    std::atomic<uint32_t> orbit_trailer_error_count;
    std::atomic<uint64_t> packet_count;
    std::atomic<uint64_t> excess_orbits_per_packet_count;
    std::atomic<uint64_t> packet_skipped_inconsistent_size;
    std::atomic<uint64_t> n_consistent_sized_packets;
  } stats;
};

#endif
